@@ -139,9 +139,9 @@ struct ao_functions {
 void ao_read_config_files (ao_config *config);
 
 #define adebug(format, args...) {\
-    if(device->verbose==2){                                             \
+    if(!device || device->verbose==2){                                  \
       if(strcmp(format,"\n")){                                          \
-        if(device->funcs->driver_info()->short_name){                   \
+        if(device && device->funcs->driver_info()->short_name){         \
           fprintf(stderr,"ao_%s debug: " format,device->funcs->driver_info()->short_name,## args); \
         }else{                                                          \
           fprintf(stderr,"debug: " format,## args);                     \
@@ -153,9 +153,9 @@ void ao_read_config_files (ao_config *config);
   }
 
 #define averbose(format, args...) {\
-    if(device->verbose>0){                                              \
+    if(!device || device->verbose>0){                                   \
       if(strcmp(format,"\n")){                                          \
-        if(device->funcs->driver_info()->short_name){                   \
+        if(device && device->funcs->driver_info()->short_name){         \
           fprintf(stderr,"ao_%s info: " format,device->funcs->driver_info()->short_name,## args); \
         }else{                                                          \
           fprintf(stderr,"info: " format,## args);                      \
@@ -167,9 +167,9 @@ void ao_read_config_files (ao_config *config);
   }
 
 #define ainfo(format, args...) {\
-    if(device->verbose>=0){                                             \
+    if(!device || device->verbose>=0){                                  \
       if(strcmp(format,"\n")){                                          \
-        if(device->funcs->driver_info()->short_name){                   \
+        if(device && device->funcs->driver_info()->short_name){         \
           fprintf(stderr,"ao_%s info: " format,device->funcs->driver_info()->short_name,## args); \
         }else{                                                          \
           fprintf(stderr,"info: " format,## args);                      \
@@ -181,9 +181,9 @@ void ao_read_config_files (ao_config *config);
   }
 
 #define awarn(format, args...) {\
-    if(device->verbose>=0){                                             \
+    if(!device || device->verbose>=0){                                  \
       if(strcmp(format,"\n")){                                          \
-        if(device->funcs->driver_info()->short_name){                   \
+        if(device && device->funcs->driver_info()->short_name){         \
           fprintf(stderr,"ao_%s WARNING: " format,device->funcs->driver_info()->short_name,## args); \
         }else{                                                          \
           fprintf(stderr,"WARNING: " format,## args);                   \
@@ -195,9 +195,9 @@ void ao_read_config_files (ao_config *config);
   }
 
 #define aerror(format, args...) {                                       \
-    if(device->verbose>=0){                                             \
+    if(!device || device->verbose>=0){                                  \
       if(strcmp(format,"\n")){                                          \
-        if(device->funcs->driver_info()->short_name){                   \
+        if(device && device->funcs->driver_info()->short_name){         \
           fprintf(stderr,"ao_%s ERROR: " format,device->funcs->driver_info()->short_name,## args); \
         }else{                                                          \
           fprintf(stderr,"ERROR: " format,## args);                     \
