@@ -23,7 +23,7 @@
  *
  ********************************************************************
 
- last mod: $Id: ao_raw.c 16982 2010-03-18 19:50:04Z xiphmont $
+ last mod: $Id$
 
  ********************************************************************/
 
@@ -44,7 +44,7 @@ static ao_info ao_raw_info =
 	AO_FMT_NATIVE,
 	0,
 	ao_raw_options,
-	5
+        sizeof(ao_raw_options)/sizeof(*ao_raw_options)
 };
 
 typedef struct ao_raw_internal
@@ -144,6 +144,12 @@ static void ao_raw_device_clear(ao_device *device)
 	ao_raw_internal *internal = (ao_raw_internal *) device->internal;
 
 	free(internal);
+        device->internal=NULL;
+}
+
+const char *ao_raw_file_extension(void)
+{
+	return "raw";
 }
 
 
@@ -155,5 +161,6 @@ ao_functions ao_raw = {
 	ao_raw_open,
 	ao_raw_play,
 	ao_raw_close,
-	ao_raw_device_clear
+	ao_raw_device_clear,
+	ao_raw_file_extension
 };
