@@ -4,6 +4,7 @@
  *
  *      Original Copyright (C) Aaron Holtzman - May 1999
  *      Modifications Copyright (C) Stan Seibert - July 2000
+ *      Modifications Copyright (C) Philipp Schafft - February 2012
  *
  *  This file is part of libao, a cross-platform audio output library.  See
  *  README for a history of this source code.
@@ -24,7 +25,7 @@
  *
  ********************************************************************
 
- last mod: $Id: audio_out.c 17008 2010-03-24 03:12:16Z xiphmont $
+ last mod: $Id$
 
  ********************************************************************/
 
@@ -633,6 +634,10 @@ static char *_sanitize_matrix(int maxchannels, char *matrix, ao_device *device){
     char *ret = calloc(strlen(matrix)+1,1); /* can only get smaller */
     char *p=matrix;
     int count=0;
+
+    if(!ret)
+      return NULL;
+
     while(count<maxchannels){
       char *h,*t;
       int m=0;
@@ -661,9 +666,8 @@ static char *_sanitize_matrix(int maxchannels, char *matrix, ao_device *device){
         /* unrecognized channel mnemonic */
         {
           int i;
-          aerror("Unrecognized channel name \"");
-          for(i=0;i<t-p;i++)fputc(p[i],stderr);
-          fprintf(stderr,"\" in channel matrix \"%s\"\n",matrix);
+          aerror("Unrecognized channel name \"%.*s\" in channel matrix \"%s\"\n",
+              t-p, p, matrix);
         }
         free(ret);
         return NULL;
@@ -706,6 +710,15 @@ static int _find_channel(int needle, char *haystack){
   return -1;
 }
 
+static void _free_map(char **m){
+  char **in=m;
+  while(m && *m){
+    free(*m);
+    m++;
+  }
+  if(in)free(in);
+}
+
 static char **_tokenize_matrix(char *matrix){
   char **ret=NULL;
   char *p=matrix;
@@ -730,6 +743,8 @@ static char **_tokenize_matrix(char *matrix){
   }
 
   ret = calloc(count+1,sizeof(*ret));
+  if(!ret)
+    return NULL;
 
   p=matrix;
   count=0;
@@ -748,6 +763,10 @@ static char **_tokenize_matrix(char *matrix){
     while(t>p && isspace(*(t-1)))t--;
 
     ret[count] = calloc(t-p+1,1);
+    if(!ret[count]){
+      _free_map(ret);
+      return NULL;
+    }
     memcpy(ret[count],p,t-p);
     count++;
     if(!*h)break;
@@ -755,16 +774,6 @@ static char **_tokenize_matrix(char *matrix){
   }
 
   return ret;
-
-}
-
-static void _free_map(char **m){
-  char **in=m;
-  while(m && *m){
-    free(*m);
-    m++;
-  }
-  if(in)free(in);
 }
 
 static unsigned int _matrix_to_channelmask(int ch, char *matrix, char *premap, int **mout){
@@ -772,7 +781,14 @@ static unsigned int _matrix_to_channelmask(int ch, char *matrix, char *premap, i
   char *p=matrix;
   int *perm=(*mout=malloc(ch*sizeof(*mout)));
   int i;
-  char **map = _tokenize_matrix(premap);
+  char **map;
+
+  if(!perm)
+    return 0;
+
+  map = _tokenize_matrix(premap);
+  if(!map)
+    return 0;
 
   for(i=0;i<ch;i++) perm[i] = -1;
   i=0;
@@ -810,6 +826,9 @@ static char *_channelmask_to_matrix(unsigned int mask, char *premap){
   char buffer[257]={0};
   char **map = _tokenize_matrix(premap);
 
+  if(!map)
+    return NULL;
+
   while(map[m]){
     if(mask & (1<<m)){
       if(count)
@@ -849,6 +868,9 @@ static char *_matrix_intersect(char *matrix,char *premap){
   int count=0;
   char **map = _tokenize_matrix(premap);
 
+  if(!map)
+    return NULL;
+
   while(1){
     char *h=p;
     int m=0;
@@ -923,6 +945,8 @@ static int ao_device_load_options(ao_device *device, ao_option *options){
     }else{
       if (!device->funcs->set_option(device, options->key, options->value)) {
         /* Problem setting options */
+        aerror("Driver %s unable to set option %s=%s\n",
+               info_table[device->driver_id]->short_name, options->key, options->value);
         return AO_EOPENDEVICE;
       }
     }
@@ -1037,7 +1061,7 @@ static ao_device* _open_device(int driver_id, ao_sample_format *format,
                                                          device->output_matrix,
                                                          &device->input_map);
               int channels = _channelmask_bits(mask);
-              if(channels<0){
+              if(channels<=0){
                 aerror("Unable to map any channels from input matrix to output");
                 errno = AO_EBADFORMAT;
                 goto error;
@@ -1058,7 +1082,7 @@ static ao_device* _open_device(int driver_id, ao_sample_format *format,
                                                          device->output_matrix,
                                                          &device->input_map);
               int channels = _channelmask_bits(mask);
-              if(channels<0){
+              if(channels<=0){
                 aerror("Unable to map any channels from input matrix to output");
                 errno = AO_EBADFORMAT;
                 goto error;
@@ -1109,6 +1133,10 @@ static ao_device* _open_device(int driver_id, ao_sample_format *format,
             int count=0;
             device->inter_permute = calloc(device->output_channels,sizeof(int));
 
+            if (!device->inter_permute) {
+              errno = AO_EFAIL;
+              goto error;
+            }
             adebug("\n");
 
             while(count<device->output_channels){
@@ -1155,8 +1183,10 @@ static ao_device* _open_device(int driver_id, ao_sample_format *format,
                 for(i=0;i<device->output_channels;i++)
                   if(device->inter_permute[i]==j)break;
                 if(i==device->output_channels){
-                  adebug("input %d (%s)\t -> none\n",
-                         j,inch[j]);
+                  if(inch){
+                    adebug("input %d (%s)\t -> none\n",
+                           j,inch[j]);
+                  }
                   unflag=1;
                 }
               }
@@ -1195,6 +1225,7 @@ static ao_device* _open_device(int driver_id, ao_sample_format *format,
 
 	if ( (device->bytewidth>1 &&
               device->client_byte_format != device->driver_byte_format) ||
+             device->input_channels != device->output_channels ||
              device->inter_permute){
 
           result = _realloc_swap_buffer(device, DEF_SWAP_BUF_SIZE);
@@ -1264,6 +1295,9 @@ void ao_shutdown(void)
 
 	if (!driver_head) return;
 
+	free(info_table);
+	info_table = NULL;
+
 	/* unload and free all the drivers */
 	while (driver) {
 		if (driver->handle) {
@@ -1475,9 +1509,12 @@ ao_info *ao_driver_info(int driver_id)
 {
 	driver_list *driver;
 
-	if ( (driver = _get_driver(driver_id)) )
-		return driver->functions->driver_info();
-	else
+	if ( (driver = _get_driver(driver_id)) ) {
+		if (driver->functions->driver_info != NULL)
+			return driver->functions->driver_info();
+		else
+			return NULL;
+	} else
 		return NULL;
 }
 
@@ -1488,6 +1525,18 @@ ao_info **ao_driver_info_list(int *count)
 	return info_table;
 }
 
+const char *ao_file_extension(int driver_id)
+{
+	driver_list *driver;
+
+	if ( (driver = _get_driver(driver_id)) ) {
+		if (driver->functions->file_extension != NULL)
+			return driver->functions->file_extension();
+		else
+			return NULL;
+	} else
+		return NULL;
+}
 
 /* -- Miscellaneous -- */
 
