@@ -25,7 +25,7 @@
  *
  ********************************************************************
 
- last mod: $Id: ao_macosx.c 18881 2013-03-14 08:44:55Z xiphmont $
+ last mod: $Id$
 
  ********************************************************************/
 /* The MacOS X CoreAudio framework doesn't mesh as simply as some
@@ -258,6 +258,10 @@ static AudioDeviceID findAudioOutputDevice(const char *name)
 {
   OSStatus err;
   AudioDeviceID aid;
+  AudioObjectPropertyAddress propertyAddress;
+  propertyAddress.mSelector = kAudioHardwarePropertyDefaultInputDevice;
+  propertyAddress.mScope = kAudioObjectPropertyScopeGlobal;
+  propertyAddress.mElement = kAudioObjectPropertyElementMaster;
 
   /* First see if it's a valid device UID */
   {
@@ -269,7 +273,7 @@ static AudioDeviceID findAudioOutputDevice(const char *name)
                                               kCFAllocatorNull);
     if (!namestr)
       return kAudioObjectUnknown;
-    err = AudioHardwareGetProperty(kAudioHardwarePropertyDeviceForUID, &size, &avt);
+    err = AudioObjectGetPropertyData(kAudioObjectSystemObject, &propertyAddress, 0, NULL, &size, &avt);
     CFRelease(namestr);
     if (!err && aid != kAudioObjectUnknown)
       return isAudioOutputDevice(aid) ? aid : kAudioObjectUnknown;
@@ -285,7 +289,7 @@ static AudioDeviceID findAudioOutputDevice(const char *name)
     if (!lcname)
       return kAudioObjectUnknown; /* no memory */
     lowercasestr(lcname);
-    err = AudioHardwareGetPropertyInfo(kAudioHardwarePropertyDevices, &size, NULL);
+    err = AudioObjectGetPropertyData(kAudioObjectSystemObject, &propertyAddress, 0, NULL, &size, devices);
     if (err) {
       free(lcname);
       return kAudioObjectUnknown;
@@ -295,7 +299,7 @@ static AudioDeviceID findAudioOutputDevice(const char *name)
       free(lcname);
       return kAudioObjectUnknown; /* no memory */
     }
-    err = AudioHardwareGetProperty(kAudioHardwarePropertyDevices, &size, devices);
+    err = AudioObjectGetPropertyData(kAudioObjectSystemObject, &propertyAddress, 0, NULL, &size, devices);
     if (err) {
       free(lcname);
       free(devices);
@@ -335,15 +339,13 @@ static AudioDeviceID findAudioOutputDevice(const char *name)
       }
       /* Check the source name too */
       size = sizeof(srcnum);
-      err = AudioDeviceGetProperty(devices[i], 0, FALSE, kAudioDevicePropertyDataSource,
-                                   &size, &srcnum);
+      AudioObjectPropertyAddress address = { kAudioDevicePropertyVolumeScalar, kAudioDevicePropertyScopeInput, 0 };
+      err = AudioObjectGetPropertyData(devices[i], &address, 0, NULL, &size, &srcnum);
       if (!err) {
         CFStringRef srcstr;
         AudioValueTranslation avt = {&srcnum, sizeof(srcnum), &srcstr, sizeof(srcstr)};
         size = sizeof(avt);
-        err = AudioDeviceGetProperty(devices[i], 0, FALSE,
-                                     kAudioDevicePropertyDataSourceNameForIDCFString,
-                                     &size, &avt);
+        err = AudioObjectGetPropertyData(devices[i], &address, 0, NULL, &size, &avt);
         if (!err && srcstr) {
           srcname = cfstringdupe(srcstr);
           CFRelease(srcstr);
@@ -408,8 +410,8 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format)
 {
   ao_macosx_internal *internal = (ao_macosx_internal *) device->internal;
   OSStatus result = noErr;
-  Component comp;
-  ComponentDescription desc;
+  AudioComponent comp;
+  AudioComponentDescription desc;
   AudioStreamBasicDescription requestedDesc;
   AURenderCallbackStruct      input;
   UInt32 i_param_size, requestedEndian;
@@ -421,14 +423,14 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format)
   desc.componentFlags = 0;
   desc.componentFlagsMask = 0;
 
-  comp = FindNextComponent (NULL, &desc);
+  comp = AudioComponentFindNext (NULL, &desc);
   if (comp == NULL) {
     aerror("Failed to start CoreAudio: AudioComponentFindNext returned NULL");
     return 0;
   }
 
   /* Open & initialize the default output audio unit */
-  result = OpenAComponent (comp, &internal->outputAudioUnit);
+  result = AudioComponentInstanceNew (comp, &internal->outputAudioUnit);
   if (result) {
     aerror("AudioComponentInstanceNew() error => %d\n",(int)result);
     return 0;
@@ -443,7 +445,7 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format)
                                    sizeof(internal->outputDevice));
     if (result) {
       aerror("AudioComponentSetDevice() error => %d\n",(int)result);
-      CloseComponent(internal->outputAudioUnit);
+      AudioComponentInstanceDispose(internal->outputAudioUnit);
       return 0;
     }
   }
@@ -592,11 +594,11 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format)
   internal->firstValidByteOffset = 0;
   internal->validByteCount = 0;
   internal->buffer = malloc(internal->bufferByteCount);
-  memset(internal->buffer, 0, internal->bufferByteCount);
   if (!internal->buffer) {
     aerror("Unable to allocate queue buffer.\n");
     return 0;
   }
+  memset(internal->buffer, 0, internal->bufferByteCount);
 
   /* limited to stereo for now */
   //if(!device->output_matrix)
@@ -723,9 +725,9 @@ int ao_plugin_close(ao_device *device)
         return 0;
       }
 
-      status = CloseComponent(internal->outputAudioUnit);
+      status = AudioComponentInstanceDispose(internal->outputAudioUnit);
       if (status) {
-        awarn("CloseComponent returned %d\n", (int)status);
+        awarn("AudioComponentInstanceDispose returned %d\n", (int)status);
         return 0;
       }
     }else
