@@ -41,7 +41,8 @@
 #define AO_NAS_BUF_SIZE 4096
 
 static char *ao_nas_options[] = {
-  "host",    /* NAS server. See nas(1) for format. */
+  "server",    /* NAS server. See nas(1) for format. */
+  "host",    /* synonym for server */
   "buf_size", /* Buffer size on server */
   "quiet",
   "verbose",
@@ -59,7 +60,7 @@ static ao_info ao_nas_info =
 	AO_FMT_NATIVE,
 	10,
 	ao_nas_options,
-	6
+        sizeof(ao_nas_options)/sizeof(*ao_nas_options)
 };
 
 
@@ -116,7 +117,7 @@ int ao_plugin_set_option(ao_device *device, const char *key, const char *value)
 {
 	ao_nas_internal *internal = (ao_nas_internal *) device->internal;
 
-	if (!strcmp(key, "host")) {
+	if (!strcmp(key, "host") || !strcmp(key, "server")) {
           char *tmp = strdup (value);
  	  if (!tmp) return 0;
  	  if (internal->host) free (internal->host);
@@ -246,8 +247,10 @@ int ao_plugin_close(ao_device *device)
 {
 	ao_nas_internal *internal = (ao_nas_internal *) device->internal;
 
-	AuStopFlow(internal->aud, internal->flow, 0);
-	AuCloseServer(internal->aud);
+        if(internal->aud){
+          AuStopFlow(internal->aud, internal->flow, 0);
+          AuCloseServer(internal->aud);
+        }
 	return 1;
 }
 
@@ -256,6 +259,8 @@ void ao_plugin_device_clear(ao_device *device)
 {
 	ao_nas_internal *internal = (ao_nas_internal *) device->internal;
 
-	free(internal->host);
+        if(internal->host)
+          free(internal->host);
 	free(internal);
+        device->internal=NULL;
 }
