@@ -18,7 +18,7 @@
 
  ********************************************************************
 
- last mod: $Id: ao_pulse.c 17007 2010-03-24 02:56:04Z xiphmont $
+ last mod: $Id$
 
  ********************************************************************/
 
@@ -30,6 +30,7 @@
 #include <assert.h>
 #include <string.h>
 #include <signal.h>
+#include <time.h>
 #include <limits.h>
 
 #include <pulse/pulseaudio.h>
@@ -40,14 +41,20 @@
 #include <ao/ao.h>
 #include <ao/plugin.h>
 
+#define AO_PULSE_BUFFER_TIME 20000
+
 /* Unfortunately libao doesn't allow "const" for these structures... */
 static char * ao_pulse_options[] = {
     "server",
     "sink",
+    "dev",
+    "id",
     "verbose",
     "quiet",
     "matrix",
-    "debug"
+    "debug",
+    "client_name"
+    "buffer_time"
 };
 
 static ao_info ao_pulse_info = {
@@ -59,12 +66,14 @@ static ao_info ao_pulse_info = {
     AO_FMT_NATIVE,
     50,
     ao_pulse_options,
-    6
+    sizeof(ao_pulse_options)/sizeof(*ao_pulse_options)
 };
 
 typedef struct ao_pulse_internal {
     struct pa_simple *simple;
-    char *server, *sink;
+    char *server, *sink, *client_name;
+    pa_usec_t static_delay;
+    pa_usec_t buffer_time;
 } ao_pulse_internal;
 
 /* Yes, this is very ugly, but required nonetheless... */
@@ -138,6 +147,8 @@ int ao_plugin_device_init(ao_device *device) {
     internal->simple = NULL;
     internal->server = NULL;
     internal->sink = NULL;
+    internal->client_name = NULL;
+    internal->buffer_time = AO_PULSE_BUFFER_TIME;
 
     device->internal = internal;
     device->output_matrix_order = AO_OUTPUT_MATRIX_PERMUTABLE;
@@ -157,11 +168,15 @@ int ao_plugin_set_option(ao_device *device, const char *key, const char *value)
     if (!strcmp(key, "server")) {
         free(internal->server);
         internal->server = strdup(value);
-    } else if (!strcmp(key, "sink")) {
+    } else if (!strcmp(key, "sink") || !strcmp(key, "dev") || !strcmp(key, "id")) {
         free(internal->sink);
         internal->sink = strdup(value);
-    } else
-        return 0;
+    } else if (!strcmp(key, "client_name")) {
+        free(internal->client_name);
+        internal->client_name = strdup(value);
+    }else if (!strcmp(key, "buffer_time")){
+      internal->buffer_time = atoi(value) * 1000;
+    }
 
     return 1;
 }
@@ -171,6 +186,7 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format) {
     ao_pulse_internal *internal;
     struct pa_sample_spec ss;
     struct pa_channel_map map;
+    struct pa_buffer_attr battr;
     size_t allocated = 128;
 
     assert(device && device->internal && format);
@@ -178,9 +194,13 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format) {
     internal = (ao_pulse_internal *) device->internal;
 
     if (format->bits == 8)
-        ss.format = PA_SAMPLE_U8;
+      ss.format = PA_SAMPLE_U8;
     else if (format->bits == 16)
-        ss.format = PA_SAMPLE_S16NE;
+      ss.format = PA_SAMPLE_S16NE;
+#ifdef PA_SAMPLE_S24NE
+    else if (format->bits == 24)
+      ss.format = PA_SAMPLE_S24NE;
+#endif
     else
         return 0;
 
@@ -192,25 +212,34 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format) {
 
     disable_sigpipe();
 
-    while (1) {
+    if (internal->client_name) {
+        snprintf(t, sizeof(t), "libao[%s]", internal->client_name);
+        snprintf(t2, sizeof(t2), "libao[%s] playback stream", internal->client_name);
+    } else {
+      while (1) {
         p = pa_xmalloc(allocated);
 
         if (!(fn = pa_get_binary_name(p, allocated))) {
-            break;
+          break;
         }
 
         if (fn != p || strlen(p) < allocated - 1) {
-            fn = pa_path_get_filename(fn);
-            snprintf(t, sizeof(t), "libao[%s]", fn);
-            snprintf(t2, sizeof(t2), "libao[%s] playback stream", fn);
-            break;
+          fn = pa_path_get_filename(fn);
+          snprintf(t, sizeof(t), "libao[%s]", fn);
+          snprintf(t2, sizeof(t2), "libao[%s] playback stream", fn);
+          break;
         }
 
         pa_xfree(p);
         allocated *= 2;
+      }
+      pa_xfree(p);
+      p = NULL;
+      if (!fn) {
+        strcpy(t, "libao");
+        strcpy(t2, "libao playback stream");
+      }
     }
-    pa_xfree(p);
-    p = NULL;
 
     if(device->input_map){
       int i;
@@ -226,12 +255,25 @@ int ao_plugin_open(ao_device *device, ao_sample_format *format) {
       }
     }
 
+    /* buffering attributes */
+    battr.prebuf = battr.minreq = battr.fragsize = -1;
 
-    if (!(internal->simple = pa_simple_new(internal->server, fn ? t : "libao", PA_STREAM_PLAYBACK, internal->sink, fn ? t2 : "libao playback stream", &ss, &map, NULL, NULL)))
+    battr.tlength = (int)(internal->buffer_time * format->rate) / 1000000 *
+      ((format->bits+7)/8) * device->output_channels;
+    battr.minreq = battr.tlength/4;
+    battr.maxlength = battr.tlength+battr.minreq;
+
+    internal->simple = pa_simple_new(internal->server, t, PA_STREAM_PLAYBACK,
+                                     internal->sink, t2, &ss,
+                                     (device->input_map ? &map : NULL), &battr, NULL);
+    if (!internal->simple)
         return 0;
 
     device->driver_byte_format = AO_FMT_NATIVE;
 
+    internal->static_delay = pa_simple_get_latency(internal->simple, NULL);
+    if(internal->static_delay<0) internal->static_delay = 0;
+
     return 1;
 }
 
@@ -247,9 +289,36 @@ int ao_plugin_close(ao_device *device) {
     assert(device && device->internal);
     ao_pulse_internal *internal = (ao_pulse_internal *) device->internal;
 
-    pa_simple_drain(internal->simple, NULL);
-    pa_simple_free(internal->simple);
-    internal->simple = NULL;
+    if(internal->simple){
+
+      /* this is a PulseAudio ALSA bug workaround;
+         pa_simple_drain() always takes about 2 seconds, even if
+         there's nothing to drain.  Rather than wait for no
+         reason, determine the current playback depth, wait
+         that long, then kill the stream.  Remove this code
+         once Pulse gets fixed. */
+
+      pa_usec_t us = pa_simple_get_latency(internal->simple, NULL);
+      if(us<0 || us>1000000){
+        pa_simple_drain(internal->simple, NULL);
+      }else{
+        us -= internal->static_delay;
+        if(us>0){
+          struct timespec sleep,wake;
+          sleep.tv_sec = (int)(us/1000000);
+          sleep.tv_nsec = (us-sleep.tv_sec*1000000)*1000;
+          while(nanosleep(&sleep,&wake)<0){
+            if(errno==EINTR)
+              sleep=wake;
+            else
+              break;
+          }
+        }
+      }
+
+      pa_simple_free(internal->simple);
+      internal->simple = NULL;
+    }
 
     return 1;
 }
@@ -258,8 +327,12 @@ void ao_plugin_device_clear(ao_device *device) {
     assert(device && device->internal);
     ao_pulse_internal *internal = (ao_pulse_internal *) device->internal;
 
-    free(internal->server);
-    free(internal->sink);
+    if(internal->server)
+      free(internal->server);
+    if(internal->sink)
+      free(internal->sink);
+    if(internal->client_name)
+      free(internal->client_name);
     free(internal);
     device->internal = NULL;
 }
